MODULE mod_etat_saha
! purpose of module is working with SAHA-S EOS tables
! in sence of CESAM architecture

    PRIVATE
! variables describe structure of table    
    INTEGER, PARAMETER :: mx=9, nt=175, nqs=100, mv=11, mv2=2  ! describe the structure of module
    
    REAL*8, DIMENSION(:,:,:,:), pointer :: xz, Beos
    REAL*8, DIMENSION(nqs,nt,mv,mx) :: Beos1, Beos2
    REAL*8, DIMENSION(mx,mv,nt,nqs) :: xz1,xz2
    TARGET xz1,xz2, Beos1, Beos2
    REAL*8, SAVE :: Ztab1,Ztab2
    
    REAL*8, PARAMETER :: Xmin=0.1d0, stepX=0.1d0
    REAL*8, DIMENSION(mx) :: Xkn=Xmin+(/(I, I=0, mx-1)/)*stepX
    REAL*8, PARAMETER :: lgTmin=3.199999947300000d0, stepT=0.025000000378736d0
    REAL*8, DIMENSION(nt) :: lgTkn=lgTmin+(/(I, I=0, nt-1)/)*stepT
    REAL*8, PARAMETER :: lgQmin=-4.50d0, stepQ=0.04999999999991d0
    REAL*8, DIMENSION(nqs) :: lgQskn=lgQmin+(/(I, I=0, nqs-1)/)*stepQ
    
    INTEGER, PARAMETER :: Jspline=4
    REAL*8, DIMENSION(mx+Jspline) :: Xknot
    REAL*8, DIMENSION(nt+Jspline) :: Tknot
    REAL*8, DIMENSION(nqs+Jspline) :: Qsknot
    
    REAL*8, PARAMETER ::  tol_extr=1.d-2    ! tolerance for extrapolation
    INTEGER, SAVE :: itime=1  
    REAL*8, DIMENSION (nqs,nt,mx) :: bb     ! coef of spline
    
    REAL*8, SAVE :: aradia=7.565912199839849d-15	!constante de la radiation
    LOGICAL, SAVE :: init=.TRUE.
    
    REAL*8, DIMENSION(:, :), ALLOCATABLE :: a  ! working region, used in bsplvd8.f
    
    PUBLIC :: etat_saha
    
CONTAINS

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       SUBROUTINE etat_saha(pp,tt,H,Z,deriv, & 
       ro,drop,drot,drox,u,dup,dut,dux,      & 
       delta,deltap,deltat,deltax,cp,dcpp,dcpt,dcpx,  &
       gradad,dgradadp,dgradadt,dgradadx,alfa,beta,gamma1)

! routine private du module mod_etat
	
! Interface of the equation of state SAHA-S3 with CESAM2k.
! Computations are performed for current metallicity Z(r,t).
! 
! input :
!	p : pressure
!	t : temperature
!	H, Z : chemical composition
!	deriv=.FALSE. : avoide computations of some derivatives

! output :
!	ro : density and derivatives
!	u : internal energy and derivatives
!   ...
 
        
      IMPLICIT NONE
      	      
      REAL*8, INTENT(in) :: H, Z
	  REAL*8, INTENT(in) :: pp, tt	
	  LOGICAL, INTENT(in) :: deriv
	
	  REAL*8, INTENT(out) :: ro,drop,drot,drox,u,dup,dut,dux
	  REAL*8, INTENT(out) :: delta,deltap,deltat,deltax,cp,dcpp,dcpt,dcpx
	  REAL*8, INTENT(out) :: gradad,dgradadp,dgradadt,dgradadx,alfa,beta, gamma1
      
      REAL*8 ro1,drop1,drot1,drox1,u1,dup1,dut1,dux1
	  REAL*8 delta1,deltap1,deltat1,deltax1,cp1,dcpp1,dcpt1,dcpx1
	  REAL*8 gradad1,dgradadp1,dgradadt1,dgradadx1,alfa1,beta1, gamma1_1
      
      REAL*8 ro2,drop2,drot2,drox2,u2,dup2,dut2,dux2
	  REAL*8 delta2,deltap2,deltat2,deltax2,cp2,dcpp2,dcpt2,dcpx2
	  REAL*8 gradad2,dgradadp2,dgradadt2,dgradadx2,alfa2,beta2, gamma1_2

      real*8 tx2, x, lgT
      REAL*8, DIMENSION(23) :: Y_var,Y_var1,Y_var2
            
   
! ________________________________________________________________________________
 
!    Range checking
     IF (Z<0.00d0.OR.Z>0.02d0) THEN
       PRINT*, 'Metallicity Z is out of the range of the SAHA-S3 tables'
       PRINT*, 'Z = ', z
       PRINT*, 'Zmin = 0.00,   Zmax = 0.02'
       STOP
     END IF  

     x=H
IF (x<0.1d0.OR.x>0.9d0) THEN
       PRINT*, 'Hydrogen mass fraction X is out of the range of the SAHA-S3 tables'
       PRINT*, 'X = ', x
       PRINT*, 'Xmin = 0.1, Xmax = 0.9'
       STOP
     END IF 

     lgT=log10(tt)
     IF (lgT<3.20d0.OR.lgT>7.55d0) THEN
       PRINT*, 'Temperature T is out of the range of the SAHA-S3 tables'
       PRINT*, 'lg(T) = ', lgT
       PRINT*, 'lg(Tmin) = 3.20, lg(Tmax) = 7.55 '
       STOP
     END IF 
     
!    Reading the SAHA_S3 tables for two z's
     IF(itime /= 12345678)THEN
        itime=12345678
        PRINT*, '------ Equation of state SAHA-S3 ------'
        CALL Z_read2_saha
     END IF  
     
    
   
!   Computing derivatives for Z=Ztab1
    xz=>xz1
    Beos=>Beos1
    CALL derivatives(pp,tt,H,deriv, & 
       ro1,drop1,drot1,drox1,u1,dup1,dut1,dux1,      & 
       delta1,deltap1,deltat1,deltax1,cp1,dcpp1,dcpt1,dcpx1,  &
       gradad1,dgradadp1,dgradadt1,dgradadx1,alfa1,beta1,gamma1_1)
    
       Y_var1=(/ro1,drop1,drot1,drox1,u1,dup1,dut1,dux1,      & 
       delta1,deltap1,deltat1,deltax1,cp1,dcpp1,dcpt1,dcpx1,  &
       gradad1,dgradadp1,dgradadt1,dgradadx1,alfa1,beta1,gamma1_1/)
       
!   Computing derivatives for Z=Ztab2
     xz=>xz2
     Beos=>Beos2
     CALL derivatives(pp,tt,H,deriv, &
       ro2,drop2,drot2,drox2,u2,dup2,dut2,dux2,      & 
       delta2,deltap2,deltat2,deltax2,cp2,dcpp2,dcpt2,dcpx2,  &
       gradad2,dgradadp2,dgradadt2,dgradadx2,alfa2,beta2,gamma1_2)
     
       Y_var2=(/ro2,drop2,drot2,drox2,u2,dup2,dut2,dux2,      & 
       delta2,deltap2,deltat2,deltax2,cp2,dcpp2,dcpt2,dcpx2,  &
       gradad2,dgradadp2,dgradadt2,dgradadx2,alfa2,beta2,gamma1_2/)
     
!   Interpolating derivatives for given metallicity Z
       tx2=(Z-Ztab1)/(Ztab2-Ztab1)
       Y_var=(1-tx2)*Y_var1+tx2*Y_var2     
    
       ro=Y_var(1)
       drop=Y_var(2)
       drot=Y_var(3)
       drox=Y_var(4)
       
       u=Y_var(5)
       dup=Y_var(6)
       dut=Y_var(7)
       dux=Y_var(8)
       
       delta=Y_var(9)
       deltap=Y_var(10)
       deltat=Y_var(11)
       deltax=Y_var(12)
       
       cp=Y_var(13)
       dcpp=Y_var(14)
       dcpt=Y_var(15)
       dcpx=Y_var(16)
       
       gradad=Y_var(17)
       dgradadp=Y_var(18)
       dgradadt=Y_var(19)
       dgradadx=Y_var(20)
       
       alfa=Y_var(21)
       beta=Y_var(22)
       gamma1=Y_var(23)    
		
	RETURN
       END SUBROUTINE etat_saha
 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!  Z_read2_saha.f90 
!
!  FUNCTIONS:
!  Z_read2_saha - read SAHA-S3 data tables eos_saha_***.dat, prepared for CESAM.
!                 The result is written in xz(mx,nt,nqs,nclm);
!               - read SAHA_S3 tables of Bspline coefficients eos_saha_Bspl_***.dat.
!
!
!***********************************************************************************************
!
!  SUBROUTINE: Z_read2_saha 
!
!  sence of vars in a line of table eos_saha_***.dat
!   P, dyne/cm^2                   u, egr/g          du/drho, CGS             {du/dT}_P=cp, CGS
!   {dlgP/dlgrho}_T=chi_rho        delta             Gamma1                   grad_ad 
!   {du/dX}_(T,P)=chi_X            rho, g/cm^3       {dlgP/dlgX}_T,rho           
!     
!***********************************************************************************************

    SUBROUTINE Z_read2_saha
    
    implicit none

    ! Variables
        
    INTEGER ix,it,iqs,imv                 ! loop indices
    
  !   Metallicity Ztab for the two SAHA-S3 tables used for Z-interpolation 
  !   and for two tables of Bspline coefficients
  !   Ztab can be 0.000, 0.010, 0.015, 0.020

    Ztab1=0.015d0
    Ztab2=0.020d0 
    
  !   Reading files with SAHA-S3 data for given metallicities Ztab1 and Ztab2
  
   PRINT*,('Reading SAHA-S data. Please, wait.')
    
   OPEN(unit=60, file='eos_saha2_015.dat', status='old', action='read')

    do ix=1,mx
     do it=1,nt
          do iqs=1,nqs
          read (60,1005) (xz1(ix,imv,it,iqs), imv=1,mv2)   
          enddo
     enddo   
    enddo
    
   close(60)
   
   OPEN(unit=61, file='eos_saha2_020.dat', status='old', action='read')
       
    do ix=1,mx
     do it=1,nt
          do iqs=1,nqs
          read (61,1005) (xz2(ix,imv,it,iqs), imv=1,mv2)  
          enddo
     enddo   
    enddo
    
    close(61)
    

  !   Reading files with tables of Bspline coefficients for given metallicities Ztab1 and Ztab2
    
     OPEN(unit=62, file='eos_saha3_Bspl_015.dat', status='old', action='read')   
    
     do ix=1,mx
       do it=1,nt
          do iqs=1,nqs
           read (62,1006) (Beos1(iqs,it,imv,ix), imv=1,mv)
          enddo
       enddo   
     enddo
    
    close(62)
   
    OPEN(unit=63, file='eos_saha3_Bspl_020.dat', status='old', action='read')
    
    do ix=1,mx
      do it=1,nt
          do iqs=1,nqs
           read (63,1006) (Beos2(iqs,it,imv,ix), imv=1,mv)  
          enddo
     enddo   
    enddo
    
    close(63)

    !  initialization B-spline knots
    call aug_knot(Xkn,Xknot)
    call aug_knot(lgTkn,Tknot)
    call aug_knot(lgQskn,Qsknot)   
    
1005      format(1x,2e18.10)     
1006      format(1x,12e25.16)
          
    END SUBROUTINE Z_read2_saha
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   SUBROUTINE int2_onlgQs(X,lgT,P_line,rho_line)
!    2D LINEAR INTERPOLATION 
!    for given X and lgT and for all net-points of lgQs 
!    of SAHA-S eos-data saved in xz array.
!    as result - P_line, rho_line.
!    essencialy, that is simplified procedure interp1
!					part of SAHA-S + CESAM project, V.Baturin, 2015, Moscow
!
    IMPLICIT NONE

    REAL*8, INTENT(in) :: X,lgT
    REAL*8, DIMENSION(nqs), intent(out) ::  P_line,rho_line  
    REAL*8, DIMENSION(2) :: tx,tt,vars_int
    INTEGER II, JJ, KK, ix, it, iq
    REAL*8, DIMENSION(mv):: Ys
    
! Check of input values (X, lgT, lgQs) for ranges 
! for belonging to permitted domain
    IF (X < 0.d0 .or. X > 1.d0) THEN
        PRINT*,'X = ', X
        STOP "SAHA_S3 \ int2_onlgQs: X is meaningless"   
    END IF    
    
   
    IF (lgT < lgTkn(1)*(1.d0-tol_extr) .or. lgT > lgTkn(nt)*(1.d0+tol_extr)) THEN
        PRINT*,'LgT = ', lgT
        PRINT*,'LgTkn(1) = ', lgTkn(1)
        PRINT*,'LgTkn(nt) = ', lgTkn(nt)
        STOP "SAHA_S3 \ int2_onlgQs: lgT is out of interval" 
    END IF

! Indexes of base point X(I), lgT(J) of interpolation cube 
    II=DINT((X-Xmin)/stepX)+1
    II=MAX(1,min(mx-1,II))
    JJ=DINT((lgT-lgTmin)/stepT)+1
    JJ=MAX(1,min(nt-1,JJ))

    DO KK=1,nqs     	
!  initialization of sum on zero is important
    vars_int=0.d0
! fractional step in every directions
    tx(2)=(X-Xkn(II))/stepX; tx(1)=(1.d0-tx(2))
    tt(2)=(lgT-lgTkn(JJ))/stepT; tt(1)=(1.d0-tt(2))

! 2D-linear interpolation
    DO ix=1,2
        DO it=1,2
			   vars_int(1)=vars_int(1)+xz(II+ix-1,2,JJ+it-1,KK)*tx(ix)*tt(it)
               vars_int(2)=vars_int(2)+xz(II+ix-1,1,JJ+it-1,KK)*tx(ix)*tt(it)
        END DO ! it
    END DO ! ix
    
     rho_line(KK)=log10(vars_int(1))
     P_line(KK)=log10(vars_int(2))

  
    END DO  ! KK
    
    RETURN
   END SUBROUTINE int2_onlgQs    
 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   
SUBROUTINE rhoof_PTX(lgT,lgP,X,eos,deos)
! give RHO as function of P, T, X according to SAHA-S3 EOS.
! suppose to be a part of module mod_etat_saha
! equivalent to rhoofp from CESAM etat_opalZ code

    IMPLICIT NONE 
    REAL*8, INTENT(in) :: lgT,lgP,X

    REAL*8, INTENT(out), DIMENSION(mv) :: eos
    REAL*8, DIMENSION(mv,3) :: deos

    REAL*8, DIMENSION(mv) :: vars
    REAL*8, DIMENSION(nqs) :: rho_line, P_line
    REAL*8 lgQs
    REAL*8 lgrho, lgrho_i, lgP_i
    INTEGER iq, iter
    
!  save all available pairs of P and rho for given T and X
	CALL int2_onlgQs(X,lgT,P_line,rho_line)

!  check this line for universality
    IF ((lgP-P_line(1)*(1.d0-SIGN(tol_extr,P_line(1))))*(lgP-P_line(nqs)*(1.d0+SIGN(tol_extr,P_line(nqs))))<=0) THEN 
       !determine rho using iterations 
       CALL interp1(P_line,rho_line,lgP,lgrho)
       lgQs=lgrho-2.25d0*(lgT-6.0d0)
       
!      Range checking
     IF (lgQs<-4.50d0.OR.lgQs>0.45d0) THEN
       PRINT*, 'Qs=rho/T6^2.25 is out of the range of the SAHA-S3 tables'
       PRINT*, 'lg(Qs) = ', lgQs
       PRINT*, 'lg(Qs_min) = -4.50, lg(Qs_max) = 0.45 '
       PRINT*, 'lg(rho) = ', lgrho
       PRINT*, 'lg(T) = ', lgT
       STOP
     END IF 

       CALL Bspline2f(x,lgT,lgQs,eos)   ! computes only eos(1) and eos(5)
       lgP_i=log10(eos(1))
       iter=1
     
       DO WHILE (ABS(lgP-lgP_i)>1d-10)
          lgrho_i=lgrho + (lgP-lgP_i)/eos(5)
          lgQs=lgrho_i-2.25d0*(lgT-6.0d0)
          
 !      Range checking
       IF (lgQs<-4.50d0.OR.lgQs>0.45d0) THEN
         PRINT*, 'Qs=rho/T6^2.25 is out of the range of the SAHA-S3 tables'
         PRINT*, 'lg(Qs) = ', lgQs
         PRINT*, 'lg(Qs_min) = -4.50, lg(Qs_max) = 0.45 '
         PRINT*, 'lg(rho) = ', lgrho
         PRINT*, 'lg(T) = ', lgT
         STOP
       END IF 
       
          CALL Bspline2f(x,lgT,lgQs,eos) ! computes only eos(1) and eos(5)
          lgP_i=log10(eos(1))
          lgrho=lgrho_i
          iter=iter+1
          IF (iter.gt.20) THEN
             PRINT*, 'SAHA-S3 \ rhoof_PTX: too many iter', lgP,lgP_i
             PAUSE
          END IF
       ENDDO
       
    ELSE
       PRINT*, 'SAHA_S3 \ rhoof_PTX: out of lgP boundary'
       PRINT*, 'P_line(1)=', P_line(1), 'P_line(nqs)=', P_line(nqs), 'lgP=',lgP
       STOP 
    END IF
    
    lgQs=lgrho-2.25d0*(lgT-6.0d0)
    
!      Range checking
    IF (lgQs<-4.50d0.OR.lgQs>0.45d0) THEN
         PRINT*, 'Qs=rho/T6^2.25 is out of the range of the SAHA-S3 tables'
         PRINT*, 'lg(Qs) = ', lgQs
         PRINT*, 'lg(Qs_min) = -4.50, lg(Qs_max) = 0.45 '
         PRINT*, 'lg(rho) = ', lgrho
         PRINT*, 'lg(T) = ', lgT
         STOP
    END IF 
    
    CALL Bspline(x,lgT,lgQs,eos,deos)   ! interpolate all functions eos and derivatives deos 

    RETURN
    END SUBROUTINE rhoof_PTX    
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   
    
    
   subroutine interp1(X_line,Y_line,X,Y)
! just 1D linear interpolation.
! It is reasonable, if the method of interpolation in accordance with interpolate
! OR alternatively this function serve as probe for PT arguments
  
    
    implicit none
    real*8, dimension(:), intent(in) :: X_line,Y_line
    real*8, intent(in) :: X
    real*8, intent(out) :: Y
   
    real*8 tx2
    integer np, ix
   
    np=SIZE(X_line);
    ix = locate(X_line,X)  ! there function for index search in arbitraty array
    ix=max(ix,1); ix=min(ix,np-1)

    tx2=(X-X_line(ix))/(X_line(ix+1)-X_line(ix))
    Y=(1-tx2)*Y_line(ix)+tx2*Y_line(ix+1) 

    return
   end subroutine interp1
   
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   
    
     SUBROUTINE derivatives(pp,tt,x,deriv, & 
       ro,drop,drot,drox,u,dup,dut,dux,      & 
       delta,deltap,deltat,deltax,cp,dcpp,dcpt,dcpx,  &
       gradad,dgradadp,dgradadt,dgradadx,alfa,beta,gamma1)

! Compute density and some other functions and their derivatives 
! for metallicity Z from the SAHA-S3 tables (Z=0.000, 0.010, 0.015, 0.020).
! 
! input :
!	p : pressure
!	t : temperature
!	x : hydrogen abundance
!	deriv=.FALSE. : avoide computations of some derivatives

! output :
!	ro : density and derivatives
!	u : internal energy and derivatives
!   ... 
    
    
      IMPLICIT NONE
    
      REAL*8, INTENT(in) :: x
	  REAL*8, INTENT(in) :: pp, tt	
	  LOGICAL, INTENT(in) :: deriv
	
	  REAL*8, INTENT(out) :: ro,drop,drot,drox,u,dup,dut,dux
	  REAL*8, INTENT(out) :: delta,deltap,deltat,deltax,cp,dcpp,dcpt,dcpx
	  REAL*8, INTENT(out) :: gradad,dgradadp,dgradadt,dgradadx,alfa,beta, gamma1
      
      REAL*8 cv, chi_T
      REAL*8 duro           
      
      REAL*8, DIMENSION(mv) :: eos
      REAL*8, DIMENSION(mv,3) :: deos  
      
      REAL*8, SAVE :: aradias3
      
      REAL*8 p, t
      REAL*8 dlg_e
! ______________________________________________________________________________  
      
      dlg_e=1.d0/LOG(10.d0)           
       
      IF(init)THEN     
	        init=.FALSE. ; aradias3=aradia/3.d0
      ENDIF	
    
      p=pp ; t=tt	

      CALL rhoof_PTX(log10(t),log10(p),x,eos,deos)
      ro=eos(10)

!   calculation TD-derivatives in point T,P,X  
	    drop=ro/p/eos(5)		   !1 / dp/dro_T
        delta=eos(6)            
        chi_T=delta*eos(5)         ! chi_T  calculated from delta 
	    drot=-ro/t*delta	       !- dp/dT / dp/dro = drho/dT_P	
	    
        u=eos(2)		   
        duro=eos(3)
	    dup=duro*drop			   !du/dro_T * dro/dp_T
        
        cp=eos(4)                  
        cv=cp-p/ro/t*chi_T*delta  
        dut=duro*drot+cv	       !du/dt_P=du/dro_T * dro/dT_P + du/dt_v 
	    gradad=eos(8)  

	    alfa=1.d0/eos(5)		   ! 1/chi ro
	    beta=1.d0-aradias3*t**4/p  
        gamma1=eos(7)
        
        drox=-eos(11)/x*ro/eos(5)  ! dro/dx_{P,T}  expressed via dP/dx_{ro,T}
        dux=eos(9)                 ! dU/dx_{P,T}
        
        IF(deriv)THEN  
        
        dcpx=deos(4,1)+deos(4,3)/ro*dlg_e*drox
        deltax=deos(6,1)+deos(6,3)/ro*dlg_e*drox
        dgradadx=deos(8,1)+deos(8,3)/ro*dlg_e*drox
        
        dcpt=(deos(4,2)-(delta+2.25d0)*deos(4,3))/T*dlg_e
        deltat=(deos(6,2)-(delta+2.25d0)*deos(6,3))/T*dlg_e
        dgradadt=(deos(8,2)-(delta+2.25d0)*deos(8,3))/T*dlg_e
        
		dcpp=deos(4,3)*drop*dlg_e/ro
		deltap=deos(6,3)*drop*dlg_e/ro
		dgradadp=deos(8,3)*drop*dlg_e/ro
        
        ELSE
         dcpx=0.d0 ; deltax= 0.d0 ; dgradadx=0.d0 ; dcpt=0.d0 ; deltat=0.d0
         dgradadt=0.d0 ; dcpp=0.d0 ; deltap=0.d0 ; dgradadp=0.d0
         
        ENDIF 	
     
    RETURN
       END SUBROUTINE derivatives
       
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       
SUBROUTINE Bspline(X0, lgT0, lgQs0, Ys, dYs)
    
    ! perform Bspline interpolation of SAHA-S3 tables for given X0, lgT0, lgQs0.
    ! Bspline computes all 11 values Ys instead of two (as Bspline2f does). 
    ! Bspline computes also derivatives dYs of some these values with respect to X, lgT, lgQs.

    IMPLICIT NONE 

    REAL*8, INTENT(in) :: X0, lgT0, lgQs0
    REAL*8, DIMENSION(mv), INTENT(out) :: Ys
    REAL*8, DIMENSION(mv,3), INTENT(out) :: dYs  ! massiv of derivatives with respect to X, lgT, lgQs

    INTEGER ix,iT,iQs,iv
    INTEGER LCX, LCT,LCQs
	INTEGER LeftX/1/, LeftT/1/, LeftQs/1/
	SAVE LeftX, LeftT, LeftQs
    REAL*8, DIMENSION(Jspline,2) :: dBIATX, dBIATT, dBIATQs
    REAL*8 bttt, bttt_X, bttt_lgT, bttt_lgQs
    INTEGER nderiv, mflag
    
    
    nderiv=2
    CALL interv8m ( Xknot, X0, LeftX, mflag )
    CALL BSPLVD8( Xknot, Jspline, X0, LeftX, dBIATX, nderiv)
    LCX=LeftX-Jspline+1
    
    CALL interv8m ( Tknot, lgT0, LeftT, mflag )
    CALL BSPLVD8( Tknot, Jspline, lgT0, LeftT, dBIATT, nderiv)
    LCT=LeftT-Jspline+1
    
    CALL interv8m ( Qsknot, lgQs0, LeftQs, mflag )
    CALL BSPLVD8( Qsknot, Jspline, lgQs0, LeftQs, dBIATQs, nderiv)
    LCQs=LeftQs-Jspline+1
        
! functions Ys   
    Ys=0.d0
    DO ix=1,Jspline
        DO iT=1,Jspline
            DO iQs=1,Jspline
                bttt=dBIATX(ix,1)*dBIATT(iT,1)*dBIATQs(iQs,1)
                DO iv=1,mv
                   Ys(iv)=Ys(iv)+Beos(LCQs+iQs-1,LCT+iT-1,iv,LCX+ix-1)*bttt
                ENDDO
            ENDDO
        ENDDO       
    ENDDO

! derivatives    
    dYs=0.d0
    DO ix=1,Jspline
       DO iT=1,Jspline
          DO iQs=1,Jspline
               bttt_X=dBIATX(ix,2)*dBIATT(iT,1)*dBIATQs(iQs,1)
               bttt_lgT=dBIATX(ix,1)*dBIATT(iT,2)*dBIATQs(iQs,1)
               bttt_lgQs=dBIATX(ix,1)*dBIATT(iT,1)*dBIATQs(iQs,2)
                DO iv=4,8,2
                   dYs(iv,1)=dYs(iv,1)+Beos(LCQs+iQs-1,LCT+iT-1,iv,LCX+ix-1)*bttt_X
                   dYs(iv,2)=dYs(iv,2)+Beos(LCQs+iQs-1,LCT+iT-1,iv,LCX+ix-1)*bttt_lgT
                   dYs(iv,3)=dYs(iv,3)+Beos(LCQs+iQs-1,LCT+iT-1,iv,LCX+ix-1)*bttt_lgQs
                ENDDO
            ENDDO
        ENDDO       
    ENDDO
          
    RETURN
END SUBROUTINE Bspline

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       
SUBROUTINE Bspline2f(X0, lgT0, lgQs0, Ys)
    
    ! perform Bspline interpolation of SAHA-S3 tables
    ! for given X0, lgT0, lgQs0.
    ! Bspline2f computes only two values Ys(1) and Ys(5)
    ! instead of all 11 (as Bspline does).
    
    IMPLICIT NONE
    
    ! Variables

    REAL*8, INTENT(in) :: X0, lgT0, lgQs0
    REAL*8, DIMENSION(mv), intent(out) :: Ys
    
    INTEGER ix,iT,iQs,iv
    INTEGER LCX, LCT,LCQs
    REAL*8, DIMENSION(Jspline) :: BIATX, BIATT, BIATQs
    REAL*8 bttt
 
    ! Body of Bspline2f 
    
    CALL BSPLVB8v( Xknot, Jspline, X0, LCX, BIATX )
    CALL BSPLVB8v( Tknot, Jspline, lgT0, LCT, BIATT )
    CALL BSPLVB8v( Qsknot, Jspline, lgQs0, LCQs, BIATQs )
   
    Ys=0.d0
    DO ix=1,Jspline
        DO iT=1,Jspline
            DO iQs=1,Jspline
                bttt=BIATX(ix)*BIATT(iT)*BIATQs(iQs)
                DO iv=1,5,4
                  Ys(iv)=Ys(iv)+Beos(LCQs+iQs-1,LCT+iT-1,iv,LCX+ix-1)*bttt
                ENDDO
            ENDDO
        ENDDO       
    ENDDO
    
    RETURN
END SUBROUTINE Bspline2f

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       
SUBROUTINE aug_knot(X,knots)
! Create knots for Bspline interpolation 
! in case of not-a-knot end condition
        
REAL*8, DIMENSION (:), INTENT(in) :: X
REAL*8, DIMENSION (:), INTENT(out) :: knots
INTEGER lenX,jk

lenX=SIZE(X)
jk=lenX+4
knots(1:4)=X(1)
knots(5:jk-4)=X(3:lenX-2)
knots(jk-3:jk)=X(lenX)

END SUBROUTINE aug_knot

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
     subroutine interv8m ( xt, x, left, mflag )
!  from  * a practical guide to splines *  by C. de Boor    
!computes  left = max( i :  xt(i) .lt. xt(lxt) .and.  xt(i) .le. x )  .
!
!******  i n p u t  ******
!  xt.....a real sequence, of length  lxt , assumed to be nondecreasing
!  lxt.....number of terms in the sequence  xt . REMOVED - used size
!  x.....the point whose location with respect to the sequence  xt  is
!        to be determined.
!
!******  o u t p u t  ******
!  left, mflag.....both integers, whose value is
!
!   1     -1      if               x .lt.  xt(1)
!   i      0      if   xt(i)  .le. x .lt. xt(i+1)
!   i      0      if   xt(i)  .lt. x .eq. xt(i+1) .eq. xt(lxt)
!   i      1      if   xt(i)  .lt.        xt(i+1) .eq. xt(lxt) .lt. x
!
!        In particular,  mflag = 0  is the 'usual' case.  mflag .ne. 0
!        indicates that  x  lies outside the CLOSED interval
!        xt(1) .le. y .le. xt(lxt) . The asymmetric treatment of the
!        intervals is due to the decision to make all pp functions cont-
!        inuous from the right, but, by returning  mflag = 0  even if
!        x = xt(lxt), there is the option of having the computed pp function
!        continuous from the left at  xt(lxt) .
!
!******  m e t h o d  ******
!  The program is designed to be efficient in the common situation that
!  it is called repeatedly, with  x  taken from an increasing or decrea-
!  sing sequence. This will happen, e.g., when a pp function is to be
!  graphed. The first guess for  left  is therefore taken to be the val-
!  ue returned at the previous call and stored in the  l o c a l  varia-
!  ble  ilo . A first check ascertains that  ilo .lt. lxt (this is nec-
!  essary since the present call may have nothing to do with the previ-
!  ous call). Then, if  xt(ilo) .le. x .lt. xt(ilo+1), we set  left =
!  ilo  and are done after just three comparisons.
!     Otherwise, we repeatedly double the difference  istep = ihi - ilo
!  while also moving  ilo  and  ihi  in the direction of  x , until
!                      xt(ilo) .le. x .lt. xt(ihi) ,
!  after which we use bisection to get, in addition, ilo+1 = ihi .
!  left = ilo  is then returned.
!
      integer, intent(inout) :: left
	  integer, intent(out) :: mflag
      integer lxt		! it was input parameter, now locals
      integer   ihi,ilo,istep,middle      ! locals
      real*8, intent(in) :: x,xt(:)
!      data ilo /1/
!      save ilo
      ilo=left		! that is value of previus call saved outside the procedure
!     it should be sure that left initially defined (where it declared)	  
	  lxt=size(xt)
      ihi = ilo + 1
      if (ihi .lt. lxt)                 go to 20
         if (x .ge. xt(lxt))            go to 110
         if (lxt .le. 1)                go to 90
         ilo = lxt - 1
         ihi = lxt
!
   20 if (x .ge. xt(ihi))               go to 40
      if (x .ge. xt(ilo))               go to 100
!
!              **** now x .lt. xt(ilo) . decrease  ilo  to capture  x .
      istep = 1
   31    ihi = ilo
         ilo = ihi - istep
         if (ilo .le. 1)                go to 35
         if (x .ge. xt(ilo))            go to 50
         istep = istep*2
                                        go to 31
   35 ilo = 1
      if (x .lt. xt(1))                 go to 90
                                        go to 50
!              **** now x .ge. xt(ihi) . increase  ihi  to capture  x .
   40 istep = 1
   41    ilo = ihi
         ihi = ilo + istep
         if (ihi .ge. lxt)              go to 45
         if (x .lt. xt(ihi))            go to 50
         istep = istep*2
                                        go to 41
   45 if (x .ge. xt(lxt))               go to 110
      ihi = lxt
!
!           **** now xt(ilo) .le. x .lt. xt(ihi) . narrow the interval.
   50 middle = (ilo + ihi)/2
      if (middle .eq. ilo)              go to 100
!     note. it is assumed that middle = ilo in case ihi = ilo+1 .
      if (x .lt. xt(middle))            go to 53
         ilo = middle
                                        go to 50
   53    ihi = middle
                                        go to 50
!**** set output and return.
   90 mflag = -1
      left = 1
                                        return
  100 mflag = 0
      left = ilo
                                        return
  110 mflag = 1
	  if (x .eq. xt(lxt)) mflag = 0
      left = lxt
  111 if (left .eq. 1)                  return
	  left = left - 1
	  if (xt(left) .lt. xt(lxt))        return
										go to 111
     end subroutine interv8m

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     
    !   original from:
!   C. de Boor. A Practical Guide To Splines. Chapter X.
!   REWRITTEN by V.Baturin, 2015
!   for purpose of not-a-knot interpolation with MATLAB
!
!   Values of all (needed) B-splines, nonzero in point X, with order Jhigh
!        simplified in respect of Index
!   knots of B-spline in array T
!
! --Input--------------------------------------------------
!
!   T - knots, length = Left+Jhigh, non-decreasing order
!
!   x - point where B-spline is computed
!
!   Left - integer, T(Left) .LE. T(Left+1)
!
! --Output-------------------------------------------------
!
!   BIATX - array with Jout length, BIATX(i) contains value
!           of polynom order Jout in point x ( B-spline
!           B(Left-Jout+i), Jout, T) on interval ( T(Left),
!           T(Left+1)
!
      SUBROUTINE BSPLVB8v( T, Jhigh, x, LC, BIATX )

 
!  Variables declaration

      IMPLICIT NONE
      
      INTEGER, INTENT(in) ::  Jhigh
      REAL*8, INTENT(in) :: T(:), x
      INTEGER, INTENT(out) ::  LC
      REAL*8, INTENT(out) :: BIATX(Jhigh)
! locals
      INTEGER Left, ki, kj, jp1
      REAL*8 DeltaL(Jhigh), DeltaR(Jhigh), Saved, Term
      INTEGER lenT, mflag
      
      kj = 1 ! order of curently calculated spline (simplification)
      lenT=SIZE(T)
!  calculation of Left (asymmetric on right edge)
      CALL interv ( T, lenT, x, Left, mflag )   
      IF (mflag.EQ.-1.OR.mflag.EQ.1) STOP 'bsplvb8v: out of interpolation range' !!!!!!!!!!
!
       LC=Left-Jhigh+1        ! LC - index in coef-array. 
!
      BIATX(1) = 1 ! value for spline of first order (constant)
      DO WHILE ( kj < Jhigh )
          jp1 = kj + 1
          DeltaR(kj) = T(Left+kj) - x
              IF (Left+1-kj.eq.0) THEN 
                PRINT*,'bsplvb8v: Left+1-kj=0' !!!!!!!!!!!!!!!!
                PRINT*,'x=',x
                PRINT*,'Left=',Left
                PRINT*,'mflag=',mflag
                PRINT*,'kj=',kj
                PRINT*,'lenT=',lenT
                STOP
              END IF
          DeltaL(kj) = x - T(Left+1-kj)
          Saved = 0.d0
!
          DO ki = 1, kj
             Term = BIATX(ki) / ( DeltaR(ki) + DeltaL(jp1-ki) )
             BIATX(ki) = Saved + DeltaR(ki)*Term
             Saved = DeltaL(jp1-ki)*Term
          ENDDO
!
         BIATX(jp1) = Saved
         kj = jp1
      ENDDO
      END SUBROUTINE BSPLVB8v

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
 subroutine bsplvd8 ( t, k, x, left, dbiatx, nderiv )
!  REWRITTEN  in doulbe precesition and fortran 90 by Baturin, 2015
!
!  from  * a practical guide to splines *  by c. de Boor (7 may 92)    
!   calls bsplvb
!   calculates value and deriv.s of all b-splines which do not vanish at x
!
!******  i n p u t  ******
!  t     the knot array, of length left+k (at least)
!  k     the order of the b-splines to be evaluated (power + 1)
!  x     the point at which these values are sought
!  left  an integer indicating the left endpoint of the interval of
!        interest. the  k  b-splines whose support contains the interval
!               (t(left), t(left+1))
!        are to be considered.
!  a s s u m p t i o n  - - -  it is assumed that
!               t(left) .lt. t(left+1)
!        division by zero will result otherwise (in  b s p l v b ).
!        also, the output is as advertised only if
!               t(left) .le. x .le. t(left+1) .
!  nderiv   an integer indicating that values of b-splines and their
!        derivatives up to but not including the  nderiv-th  are asked
!        for. ( nderiv  is replaced internally by the integer  m h i g h
!        in  (1,k)  closest to it.)
!
!******  w o r k   a r e a  ******  <just pre-allocated outside array, with dynamic dimensions>
!  a     an array of order (k,k), to contain b-coeff.s of the derivat-
!        ives of a certain order of the  k  b-splines of interest.
!
!******  o u t p u t  ******
!  dbiatx   an array of order (k,nderiv). its entry  (i,m)  contains
!        value of  (m-1)st  derivative of  (left-k+i)-th  b-spline of
!        order  k  for knot sequence  t , i=1,...,k, m=1,...,nderiv.
!
!******  m e t h o d  ******
!  values at  x  of all the relevant b-splines of order k,k-1,...,
!  k+1-nderiv  are generated via  bsplvb  and stored temporarily in
!  dbiatx .  then, the b-coeffs of the required derivatives of the b-
!  splines of interest are generated by differencing, each from the pre-
!  ceding one of lower order, and combined with the values of b-splines
!  of corresponding order in  dbiatx  to produce the desired values .
!
      implicit none
      integer, intent (in) :: k,left,nderiv
      
      real*8, intent (out) :: dbiatx(k,nderiv)
      real*8, intent (in) :: t(:),x
!  locals:
      real*8 dbiat1(k)   ! one-dimensional array, use in bsplvb call
      real*8 factor,fkp1mm,sum
      integer ideriv, mhigh
      Integer i,il,j,jlow,jp1mid,kp1,kp1mm,ldummy,m
      integer ierr
      
      mhigh = max0(min0(nderiv,k),1)
!     mhigh is usually equal to nderiv.
      kp1 = k+1
      call bsplvb8(t,kp1-mhigh,1,x,left,dbiat1)
      dbiatx(mhigh:k,mhigh) = dbiat1(1:kp1-mhigh)
      if (mhigh .eq. 1) return
!     the first column of  dbiatx  always contains the b-spline values
!     for the current order. these are stored in column k+1-current
!     order  before  bsplvb  is called to put values for the next
!     higher order on top of it.
      ideriv = mhigh
      do m=2,mhigh
!         jp1mid = 1
!         do j=ideriv,k
!            jp1mid = jp1mid + 1
!         enddo
         ideriv = ideriv - 1
         call bsplvb8(t,kp1-ideriv,2,x,left,dbiat1)
          dbiatx(ideriv:k,ideriv) = dbiat1(1:kp1-ideriv)
      enddo
!
!     at this point,  b(left-k+i, k+1-j)(x) is in  dbiatx(i,j) for
!     i=j,...,k and j=1,...,mhigh ('=' nderiv). in particular, the
!     first column of  dbiatx  is already in final form. to obtain cor-
!     responding derivatives of b-splines in subsequent columns, gene-
!     rate their b-repr. by differencing, then evaluate at  x.
!
!  Esencialy, this block is initiation of a() to indentity matrix
      if (.not.allocated(a)) then
          allocate (a(k,k), stat=ierr)
          if (ierr.ne.0) stop 'Allocation of <a> error'
      elseif (size(a,dim=1).ne.k) then
          deallocate(a)
          allocate (a(k,k), stat=ierr)
          if (ierr.ne.0) stop 'Allocation of <a> error'
      end if

       a = 0.d0                           ! Initialize the array.
       forall(j = 1:k) a(j,j) = 1.d0      ! Set the diagonal.
!
!      jlow = 1
!      do i=1,k
!         do j=jlow,k
!         a(j,i) = 0.d0
!         enddo
!         jlow = i
!      a(i,i) = 1.
!      enddo
!
!     at this point, a(.,j) contains the b-coeffs for the j-th of the
!     k  b-splines of interest here.
!
      do m=2,mhigh
         kp1mm = kp1 - m
         fkp1mm = float(kp1mm)
         il = left
         i = k
!
!        for j=1,...,k, construct b-coeffs of  (m-1)st  derivative of
!        b-splines from those for preceding derivative by differencing
!        and store again in  a(.,j) . the fact that  a(i,j) = 0  for
!        i .lt. j  is used.
         do ldummy=1,kp1mm
            factor = fkp1mm/(t(il+kp1mm) - t(il))
!           the assumption that t(left).lt.t(left+1) makes denominator
!           in  factor  nonzero.
            do j=1,i
               a(i,j) = (a(i,j) - a(i-1,j))*factor
            enddo
            il = il - 1
         i = i - 1
         enddo
!
!        for i=1,...,k, combine b-coeffs a(.,i) with b-spline values
!        stored in dbiatx(.,m) to get value of  (m-1)st  derivative of
!        i-th b-spline (of interest here) at  x , and store in
!        dbiatx(i,m). storage of this value over the value of a b-spline
!        of order m there is safe since the remaining b-spline derivat-
!        ives of the same order do not use this value due to the fact
!        that  a(j,i) = 0  for j .lt. i .
         do i=1,k
            sum = 0.d0
            jlow = max0(i,m)
            do j=jlow,k
              sum = a(j,i)*dbiatx(j,m) + sum
            enddo
         dbiatx(i,m) = sum
         enddo
         enddo
      return
      end subroutine bsplvd8
 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
    
  subroutine bsplvb8 ( t, jhigh, index, x, left, biatx )
!  from  * a practical guide to splines *  by c. de boor    
! calculates the value of all possibly nonzero b-splines at  x  of order
!
!               jout  =  max( jhigh , (j+1)*(index-1) )
!
!  with knot sequence  t .
!
!******  i n p u t  ******
!  t.....knot sequence, of length  left + jout  , assumed to be nonde-
!        creasing.  a s s u m p t i o n . . . .
!                       t(left)  .lt.  t(left + 1)   .
!   d i v i s i o n  b y  z e r o  will result if  t(left) = t(left+1)
!  jhigh,
!  index.....integers which determine the order  jout = max(jhigh,
!        (j+1)*(index-1))  of the b-splines whose values at  x  are to
!        be returned.  index  is used to avoid recalculations when seve-
!        ral columns of the triangular array of b-spline values are nee-
!        ded (e.g., in  bsplpp  or in  bsplvd ). precisely,
!                     if  index = 1 ,
!        the calculation starts from scratch and the entire triangular
!        array of b-spline values of orders 1,2,...,jhigh  is generated
!        order by order , i.e., column by column .
!                     if  index = 2 ,
!        only the b-spline values of order  j+1, j+2, ..., jout  are ge-
!        nerated, the assumption being that  biatx , j , deltal , deltar
!        are, on entry, as they were on exit at the previous call.
!           in particular, if  jhigh = 0, then  jout = j+1, i.e., just
!        the next column of b-spline values is generated.
!
!  w a r n i n g . . .  the restriction   jout .le. jmax (= 20)  is im-
!        posed arbitrarily by the dimension statement for  deltal  and
!        deltar  below, but is  n o w h e r e  c h e ! k e d  for .
!
!  x.....the point at which the b-splines are to be evaluated.
!  left.....an integer chosen (usually) so that
!                  t(left) .le. x .le. t(left+1)  .
!
!******  o u t p u t  ******
!  biatx.....array of length  jout , with  biatx(i)  containing the val-
!        ue at  x  of the polynomial of order  jout  which agrees with
!        the b-spline  b(left-jout+i,jout,t)  on the interval (t(left),
!        t(left+1)) .
!
!******  m e t h o d  ******
!  the recurrence relation
!
!                       x - t(i)              t(i+j+1) - x
!     b(i,j+1)(x)  =  -----------b(i,j)(x) + ---------------b(i+1,j)(x)
!                     t(i+j)-t(i)            t(i+j+1)-t(i+1)
!
!  is used (repeatedly) to generate the (j+1)-vector  b(left-j,j+1)(x),
!  ...,b(left,j+1)(x)  from the j-vector  b(left-j+1,j)(x),...,
!  b(left,j)(x), storing the new values in  biatx  over the old. the
!  facts that
!            b(i,1) = 1  if  t(i) .le. x .lt. t(i+1)
!  and that
!            b(i,j)(x) = 0  unless  t(i) .le. x .lt. t(i+j)
!  are used. the particular organization of the calculations follows al-
!  gorithm  (8)  in chapter x of the text.
!
      integer, intent(in) :: index, jhigh, left
      integer i, j, jmax, jp1
      parameter (jmax = 20)
      real*8 biatx(jhigh), t(:), x
      real*8 deltal(jmax), deltar(jmax), saved, term
!     real biatx(jhigh),t(1),x,   deltal(20),deltar(20),saved,term
!     dimension biatx(jout), t(left+jout)
!  current fortran standard makes it impossible to specify the length of
!  t  and of  biatx  precisely without the introduction of otherwise
!  superfluous additional arguments.
      data j/1/
      save j,deltal,deltar 
!
      go to (10,20), index
   10 j = 1
      biatx(1) = 1.d0
      if (j .ge. jhigh)       return
!
   20    jp1 = j + 1
         deltar(j) = t(left+j) - x
         deltal(j) = x - t(left+1-j)
         saved = 0.d0
         do i=1,j
            term = biatx(i)/(deltar(i) + deltal(jp1-i))
            biatx(i) = saved + deltar(i)*term
            saved = deltal(jp1-i)*term
         enddo
         biatx(jp1) = saved
         j = jp1
         if (j .lt. jhigh)              go to 20
!
      return
  end subroutine bsplvb8
 
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
FUNCTION locate(xx,x)
!  find left index of interval of xx includes x

	IMPLICIT NONE
	REAL*8, DIMENSION(:), INTENT(IN) :: xx
	REAL*8, INTENT(IN) :: x
	INTEGER locate
!  locals
	INTEGER :: n,jl,jm,ju
	LOGICAL :: ascnd
	n=size(xx)
	ascnd = (xx(n) >= xx(1)) ! check for acsendance of xx
	jl=0        ! low boundary
	ju=n+1      ! upper boundary
	do while (ju-jl > 1) 
		jm=(ju+jl)/2
		if (ascnd .eqv. (x >= xx(jm))) then
			jl=jm
		else
			ju=jm
		end if
	enddo
	if (x == xx(1)) then
		locate=1
	else if (x == xx(n)) then
		locate=n-1
	else
		locate=jl
	end if
    END FUNCTION locate
    
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine interv ( xt, lxt, x, left, mflag )
!c  from  * a practical guide to splines *  by C. de Boor    
!C computes  left = max( i :  xt(i) .lt. xt(lxt) .and.  xt(i) .le. x )  .
!c
!c******  i n p u t  ******
!c  xt.....a real sequence, of length  lxt , assumed to be nondecreasing
!c  lxt.....number of terms in the sequence  xt .
!c  x.....the point whose location with respect to the sequence  xt  is
!c        to be determined.
!c
!c******  o u t p u t  ******
!c  left, mflag.....both integers, whose value is
!c
!c   1     -1      if               x .lt.  xt(1)
!c   i      0      if   xt(i)  .le. x .lt. xt(i+1)
!c   i      0      if   xt(i)  .lt. x .eq. xt(i+1) .eq. xt(lxt)
!c   i      1      if   xt(i)  .lt.        xt(i+1) .eq. xt(lxt) .lt. x
!c
!c        In particular,  mflag = 0  is the 'usual' case.  mflag .ne. 0
!c        indicates that  x  lies outside the CLOSED interval
!c        xt(1) .le. y .le. xt(lxt) . The asymmetric treatment of the
!c        intervals is due to the decision to make all pp functions cont-
!c        inuous from the right, but, by returning  mflag = 0  even if
!C        x = xt(lxt), there is the option of having the computed pp function
!c        continuous from the left at  xt(lxt) .
!c
!c******  m e t h o d  ******
!c  The program is designed to be efficient in the common situation that
!c  it is called repeatedly, with  x  taken from an increasing or decrea-
!c  sing sequence. This will happen, e.g., when a pp function is to be
!c  graphed. The first guess for  left  is therefore taken to be the val-
!c  ue returned at the previous call and stored in the  l o c a l  varia-
!c  ble  ilo . A first check ascertains that  ilo .lt. lxt (this is nec-
!c  essary since the present call may have nothing to do with the previ-
!c  ous call). Then, if  xt(ilo) .le. x .lt. xt(ilo+1), we set  left =
!c  ilo  and are done after just three comparisons.
!c     Otherwise, we repeatedly double the difference  istep = ihi - ilo
!c  while also moving  ilo  and  ihi  in the direction of  x , until
!c                      xt(ilo) .le. x .lt. xt(ihi) ,
!c  after which we use bisection to get, in addition, ilo+1 = ihi .
!c  left = ilo  is then returned.
!c
      integer left,lxt,mflag,   ihi,ilo,istep,middle
      real*8 x,xt(lxt)
      data ilo /1/
      save ilo  
      ihi = ilo + 1
      if (ihi .lt. lxt)                 go to 20
         if (x .ge. xt(lxt))            go to 110
         if (lxt .le. 1)                go to 90
         ilo = lxt - 1
         ihi = lxt
!c
   20 if (x .ge. xt(ihi))               go to 40
      if (x .ge. xt(ilo))               go to 100
!c
!c              **** now x .lt. xt(ilo) . decrease  ilo  to capture  x .
      istep = 1
   31    ihi = ilo
         ilo = ihi - istep
         if (ilo .le. 1)                go to 35
         if (x .ge. xt(ilo))            go to 50
         istep = istep*2
                                        go to 31
   35 ilo = 1
      if (x .lt. xt(1))                 go to 90
                                        go to 50
!c              **** now x .ge. xt(ihi) . increase  ihi  to capture  x .
   40 istep = 1
   41    ilo = ihi
         ihi = ilo + istep
         if (ihi .ge. lxt)              go to 45
         if (x .lt. xt(ihi))            go to 50
         istep = istep*2
                                        go to 41
   45 if (x .ge. xt(lxt))               go to 110
      ihi = lxt
!c
!c           **** now xt(ilo) .le. x .lt. xt(ihi) . narrow the interval.
   50 middle = (ilo + ihi)/2
      if (middle .eq. ilo)              go to 100
!c     note. it is assumed that middle = ilo in case ihi = ilo+1 .
      if (x .lt. xt(middle))            go to 53
         ilo = middle
                                        go to 50
   53    ihi = middle
                                        go to 50
!c**** set output and return.
   90 mflag = -1
      left = 1
                                        return
  100 mflag = 0
      left = ilo
                                        return
  110 mflag = 1
	  if (x .eq. xt(lxt)) mflag = 0
      left = lxt
  111 if (left .eq. 1)                  return
	  left = left - 1
	  if (xt(left) .lt. xt(lxt))        return
										go to 111
    end subroutine interv
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
   
  END MODULE mod_etat_saha  
